#!/usr/bin/env python
# ct_statevector_tools.py

"""
Author : peters 

Revision History:
File created on 28 Jul 2010.

"""

import os
import sys
import logging
import datetime


################### Begin Class EnsembleMember ###################

class EnsembleMember(object):
    """ 
        An ensemble member. This consists of
           * a member number
           * parameter values
           * a ModelSample object to hold sampled values for this member
    """

    def __init__(self, membernumber):
        self.membernumber           = membernumber   # the member number
        self.ParameterValues        = None           # Parameter values of this member
        self.ModelSample            = None           # Model Sampled Parameter values of this member

    def __str__(self):
        return "%03d"%self.membernumber

    def WriteToFile(self, outdir):
        """ Write the information needed by an external model to a netcdf file for future use """
        from da.tools.io import CT_CDF, std_savedict
        import numpy as np

        filename        = os.path.join(outdir,'parameters.%03d.nc'% self.membernumber)
        f               = CT_CDF(filename,'create')
        dimparams       = f.AddParamsDim(len(self.ParameterValues))

        data =  self.ParameterValues

        savedict                = std_savedict.copy()
        savedict['name']        = "parametervalues"
        savedict['long_name']   = "parameter_values_for_member_%d"%self.membernumber
        savedict['units']       = "unitless"
        savedict['dims']        = dimparams 
        savedict['values']      = data.tolist()
        savedict['comment']     = 'These are parameter values to use for member %d'%self.membernumber
        dummy                   = f.AddData(savedict)

        dummy           = f.close()

        msg     = 'Successfully wrote data from ensemble member %d to file (%s) ' % (self.membernumber,filename,) ; logging.info(msg)



################### End Class EnsembleMember ###################

################### Begin Class StateVector ###################

import numpy as np
class StateVector(object):
    """ an object that holds data + methods and attributes needed to manipulate state vector values """
    def __init__(self,dims):

        self.nlag               = dims[0]
        self.nmembers           = dims[1]
        self.nparams            = dims[2]
        self.isOptimized        = False

        # These list objects hold the data for each time step of lag in the system. Note that the ensembles for each time step consist 
        # of lists of EnsembleMember objects, we define member 0 as the mean of the distribution and n=1,...,nmembers as the spread.

        self.EnsembleMembers    = range(self.nlag)

        for n in range(self.nlag):
            self.EnsembleMembers[n] = []

    def __str__(self):
        return "This is a base class derived state vector object"

    def MakeNewEnsemble(self,lag, covariancematrix = None):
        """ Make a new ensemble, the attribute lag refers to the position in the state vector. 
            Note that lag=1 means an index of 0 in python, hence the notation lag-1 in the indexing below.
            The argument is thus referring to the lagged state vector as [1,2,3,4,5,..., nlag]

            The optional DaCycle object to be passed holds info on the data assimilation system settings needed
            for instance to read covariance data from file. This is not part of the base class object here, hence, this
            mehtod will almost always be overwritten in the derived class for the specific application.

            The code below is just an example of what could be used
        """    

        if covariancematrix == None: 
            covariancematrix=np.identity(self.nparams)

        # Make a cholesky decomposition of the covariance matrix

        U,s,Vh  = np.linalg.svd(covariancematrix)
        dof     = np.sum(s)**2/sum(s**2)
        C       = np.linalg.cholesky(covariancematrix)

        msg =   'Cholesky decomposition has succeeded offline'                    ; logging.debug(msg)
        msg =   'Appr. degrees of freedom in covariance matrix is %s'%(int(dof))  ; logging.info(msg)


        # Create mean values 

        NewMean                     = np.ones(self.nparams,float)

        # Create the first ensemble member with a deviation of 0.0 and add to list

        NewMember                   = EnsembleMember(0)
        NewMember.ParameterValues   = np.zeros((self.nparams),float)  + NewMean
        dummy                       = self.EnsembleMembers[lag-1].append(NewMember)

        # Create members 1:nmembers and add to EnsembleMembers list

        for member in range(1,self.nmembers):
            randstate                   = np.random.get_state()
            rands                       = np.random.randn(self.nparams)

            NewMember                   = EnsembleMember(member)
            NewMember.ParameterValues   = np.dot(C,rands) + NewMean
            dummy                       = self.EnsembleMembers[lag-1].append(NewMember)

        msg =   '%d new ensemble members were added to the state vector # %d'%(self.nmembers,lag)  ; logging.debug(msg)

    def Propagate(self):
        """ 

        Propagate the parameter values in the StateVector to the next cycle. This means a shift by one cycle step for all states that will
        be optimized once more, and the creation of a new ensemble for the time step that just comes in for the first time (step=nlag). 
        In the future, this routine can incorporate a formal propagation of the statevector.

        """

        # Remove State Vector n=1 by simply "popping" it from the list and appending a new empty list at the front. This empty list will
        # hold the new ensemble for the new cycle 

        dummy = self.EnsembleMembers.pop(0)

        dummy = self.EnsembleMembers.append([])

        # And now create a new week of mean + members for n=nlag

        dummy = self.MakeNewEnsemble(self.nlag)

        msg   =   'The state vector has been propagated by one cycle ' ; logging.info(msg)

        return None

    def WriteToFile(self,filename):
        """ Write the StateVector object to a netcdf file for future use """
        from da.tools.io import CT_CDF
        import numpy as np

        f               = CT_CDF(filename,'create')
        dimparams       = f.AddParamsDim(self.nparams)
        dimmembers      = f.AddMembersDim(self.nmembers)
        dimlag          = f.AddLagDim(self.nlag,unlimited=True)

        for n in range(self.nlag):

            members = self.EnsembleMembers[n]
            MeanState =  members[0].ParameterValues

            savedict                = f.StandardVar(varname='meanstate')
            savedict['dims']        = dimlag+dimparams 
            savedict['values']      = MeanState
            savedict['count']       = n
            savedict['comment']     = 'this represents the mean of the ensemble'
            dummy                   = f.AddData(savedict)

            members =  self.EnsembleMembers[n]
            data = np.array([m.ParameterValues for m in members])- MeanState

            savedict                = f.StandardVar(varname='ensemblestate')
            savedict['dims']        = dimlag+dimmembers+dimparams 
            savedict['values']      = data
            savedict['count']       = n
            savedict['comment']     = 'this represents deviations from the mean of the ensemble'
            dummy                   = f.AddData(savedict)

        dummy           = f.close()

        msg     = 'Successfully wrote the State Vector to file (%s) ' % (filename,) ; logging.info(msg)

    def ReadFromFile(self,filename):
        """ Read the StateVector object from a netcdf file for future use """
        import Nio
        import numpy as np

        f               = Nio.open_file(filename,'r')
        MeanState       = f.variables['statevectormean'].get_value()
        EnsembleMembers = f.variables['statevectorensemble'].get_value()
        dummy           = f.close()

        for n in range(self.nlag):
            for m in range(self.nmembers):
                NewMember                   = EnsembleMember(m)
                NewMember.ParameterValues   = EnsembleMembers[n,m,:] + MeanState  # add the mean to the deviations to hold the full parameter values
                dummy                       = self.EnsembleMembers[n].append(NewMember)

        msg     = 'Successfully read the State Vector from file (%s) ' % (filename,) ; logging.info(msg)

################### End Class StateVector ###################

if __name__ == "__main__":

    sys.path.append('../../')

    import os
    import sys
    from da.tools.general import StartLogger 
    from da.tools.initexit import CycleControl 
    import numpy as np
    import da.tools.rc as rc

    opts = ['-v']
    args = {'rc':'../../da.rc','logfile':'da_initexit.log','jobrcfilename':'test.rc'}

    StartLogger()
    DaCycle = CycleControl(opts,args)

    dummy = EnsembleMember(0)
    print dummy

    DaCycle.Initialize()
    print DaCycle

    dims        = ( int(DaCycle.da_settings['time.nlag']),
                  int(DaCycle.da_settings['forecast.nmembers']),
                  int(DaCycle.DaSystem.da_settings['nparameters']),
                  )

    StateVector = StateVector(dims)

    StateVector.MakeNewEnsemble(lag=1)
    StateVector.MakeNewEnsemble(lag=2)

    members =  StateVector.EnsembleMembers[1]

    members[0].WriteToFile(DaCycle.da_settings['dir.input'])

    StateVector.Propagate()

    savedir         = DaCycle.da_settings['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    StateVector.WriteToFile(filename)


    savedir         = DaCycle.da_settings['dir.output']
    filename        = os.path.join(savedir,'savestate.nc')

    StateVector.ReadFromFile(filename)

